package benchmark;

/* 
 * Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 */
public class Main
{

	/**
	 * Returns an exact representation of the <a
	 * href="http://mathworld.wolfram.com/BinomialCoefficient.html"> Binomial Coefficient</a>, "
	 * <code>n choose k</code>", the number of <code>k</code>-element subsets that can be selected
	 * from an <code>n</code>-element set.
	 * <p>
	 * <Strong>Preconditions</strong>:
	 * <ul>
	 * <li> <code>0 <= k <= n </code> (otherwise <code>IllegalArgumentException</code> is thrown)</li>
	 * <li>The result is small enough to fit into a <code>long</code>. The largest value of
	 * <code>n</code> for which all coefficients are <code> < Long.MAX_VALUE</code> is 66. If the
	 * computed value exceeds <code>Long.MAX_VALUE</code> an <code>ArithMeticException
	 *      </code> is thrown.</li>
	 * </ul>
	 * </p>
	 * 
	 * @param n the size of the set
	 * @param k the size of the subsets to be counted
	 * @return <code>n choose k</code>
	 * @throws IllegalArgumentException if preconditions are not met.
	 * @throws ArithmeticException if the result is too large to be represented by a long integer.
	 */
	public static long binomialCoefficient(final int n, final int k)
	{
		if ( n < k )
			throw new IllegalArgumentException("must have n >= k for binomial coefficient (n,k)");
		if ( n < 0 )
			throw new IllegalArgumentException("must have n >= 0 for binomial coefficient (n,k)");
		if ( n == k || k == 0 )
			return 1;
		if ( k == 1 || k == n - 1 )
			return n;

		final long result = Math.round(Main.binomialCoefficientDouble(n, k));
		if ( result == Long.MAX_VALUE )
			throw new ArithmeticException("result too large to represent in a long integer");
		return result;
	}

	/**
	 * Returns a <code>double</code> representation of the <a
	 * href="http://mathworld.wolfram.com/BinomialCoefficient.html"> Binomial Coefficient</a>, "
	 * <code>n choose k</code>", the number of <code>k</code>-element subsets that can be selected
	 * from an <code>n</code>-element set.
	 * <p>
	 * <Strong>Preconditions</strong>:
	 * <ul>
	 * <li> <code>0 <= k <= n </code> (otherwise <code>IllegalArgumentException</code> is thrown)</li>
	 * <li>The result is small enough to fit into a <code>double</code>. The largest value of
	 * <code>n</code> for which all coefficients are < Double.MAX_VALUE is 1029. If the computed
	 * value exceeds Double.MAX_VALUE, Double.POSITIVE_INFINITY is returned</li>
	 * </ul>
	 * </p>
	 * 
	 * @param n the size of the set
	 * @param k the size of the subsets to be counted
	 * @return <code>n choose k</code>
	 * @throws IllegalArgumentException if preconditions are not met.
	 */
	public static double binomialCoefficientDouble(final int n, final int k)
	{
		return Math.floor(Math.exp(Main.binomialCoefficientLog(n, k)) + 0.5);
	}

	/**
	 * Returns the natural <code>log</code> of the <a
	 * href="http://mathworld.wolfram.com/BinomialCoefficient.html"> Binomial Coefficient</a>, "
	 * <code>n choose k</code>", the number of <code>k</code>-element subsets that can be selected
	 * from an <code>n</code>-element set.
	 * <p>
	 * <Strong>Preconditions</strong>:
	 * <ul>
	 * <li> <code>0 <= k <= n </code> (otherwise <code>IllegalArgumentException</code> is thrown)</li>
	 * </ul>
	 * </p>
	 * 
	 * @param n the size of the set
	 * @param k the size of the subsets to be counted
	 * @return <code>n choose k</code>
	 * @throws IllegalArgumentException if preconditions are not met.
	 */
	public static double binomialCoefficientLog(final int n, final int k)
	{
		if ( n < k )
			throw new IllegalArgumentException("must have n >= k for binomial coefficient (n,k)");
		if ( n < 0 )
			throw new IllegalArgumentException("must have n >= 0 for binomial coefficient (n,k)");
		if ( n == k || k == 0 )
			return 0;
		if ( k == 1 || k == n - 1 )
			return Math.log(n);
		double logSum = 0;

		// n!/k!
		for (int i = k + 1; i <= n; i++)
		{
			logSum += Math.log(i);
		}

		// divide by (n-k)!
		for (int i = 2; i <= n - k; i++)
		{
			logSum -= Math.log(i);
		}

		return logSum;
	}
}