package dataStructures;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.NoSuchElementException;

public final class BinaryHeap<V> extends java.util.AbstractQueue<V>
{

	private final static int DEFAULT_CAPACITY = 13;

	/**
	 * The elements in this heap.
	 */
	private V[] array;

	private final HashMap<V, Integer> elementsToIndexMap = new HashMap<V, Integer>();

	/**
	 * The number of elements currently in this heap.
	 */
	private int heapSize;

	/**
	 * The comparator used to order the elements
	 */
	private Comparator<? super V> m_comparator;

	/**
	 * Constructs a new minimum binary heap.
	 */
	public BinaryHeap()
	{
		this(BinaryHeap.DEFAULT_CAPACITY);
	}

	public BinaryHeap(Collection<V> vertexSet)
	{
		this(vertexSet.size());
		this.iniHeap(vertexSet);
	}

	public BinaryHeap(Comparator<? super V> comparator, Collection<V> vertexSet)
	{
		this(Math.max(vertexSet.size(), BinaryHeap.DEFAULT_CAPACITY), comparator);
		this.iniHeap(vertexSet);
	}

	/**
	 * Constructs a new <code>BinaryHeap</code> that will use the given comparator to order its
	 * elements.
	 * 
	 * @param comparator the comparator used to order the elements, null means use natural order
	 */
	public BinaryHeap(Comparator<V> comparator)
	{
		this();
		this.m_comparator = comparator;
	}

	/**
	 * Constructs a new minimum or maximum binary heap with the specified initial capacity.
	 * 
	 * @param capacity the initial capacity for the heap. This value must be greater than zero.
	 * @param isMinHeap if <code>true</code> the heap is created as a minimum heap; otherwise, the
	 *            heap is created as a maximum heap.
	 * @throws IllegalArgumentException if <code>capacity</code> is <code>&lt;= 0</code>
	 */
	public BinaryHeap(int capacity)
	{
		if ( capacity <= 0 )
			throw new IllegalArgumentException("invalid capacity");

		// +1 as 0 is noop
		this.array = (V[]) new Object[capacity + 1];
	}

	/**
	 * Constructs a new <code>BinaryHeap</code>.
	 * 
	 * @param capacity the initial capacity for the heap
	 * @param comparator the comparator used to order the elements, null means use natural order
	 * @throws IllegalArgumentException if <code>capacity</code> is &lt;= <code>0</code>
	 */
	public BinaryHeap(int capacity, Comparator<? super V> comparator)
	{
		this(capacity);
		this.m_comparator = comparator;
	}

	@Override
	public boolean add(V e)
	{
		this.insert(e);
		return true;
	}

	@Override
	public boolean addAll(Collection<? extends V> c)
	{
		for (final V v : c)
		{
			this.insert(v);
		}
		return true;
	}

	/**
	 * Clears all elements from queue.
	 */
	@Override
	public void clear()
	{
		this.array = (V[]) new Object[this.array.length]; // for gc
		this.elementsToIndexMap.clear();
		this.heapSize = 0;
	}

	/**
	 * Compares two objects using the comparator if specified, or the natural order otherwise.
	 * 
	 * @param a the first object
	 * @param b the second object
	 * @return -ve if a less than b, 0 if they are equal, +ve if a greater than b
	 */
	private int compare(V a, V b)
	{
		if ( this.m_comparator != null )
			return this.m_comparator.compare(a, b);
		else
			return ((Comparable) a).compareTo(b);
	}

	public boolean containElement(V v)
	{
		return this.elementsToIndexMap.containsKey(v);
	}

	@Override
	public boolean contains(Object o)
	{
		return this.elementsToIndexMap.containsKey(o);
	}

	public BinaryHeap<V> copy()
	{
		final BinaryHeap<V> t = new BinaryHeap<V>();
		final V[] elements = (V[]) new Object[this.array.length];
		System.arraycopy(this.array, 0, elements, 0, this.array.length);
		t.array = elements;
		t.heapSize = this.heapSize;
		t.m_comparator = this.m_comparator;
		t.elementsToIndexMap.clear();
		for (final V v : this.elementsToIndexMap.keySet())
		{
			t.elementsToIndexMap.put(v, this.elementsToIndexMap.get(v));
		}
		return t;
	}

	public void ensureHeapProperty()
	{
		for (int i = this.heapSize / 2; i >= 1; i--)
		{
			this.percolateDownMinHeap(i);
		}
	}

	public V get(V vertexTwoNeighborhood)
	{
		if ( !this.elementsToIndexMap.containsKey(vertexTwoNeighborhood) )
			throw new NoSuchElementException();
		return this.array[this.elementsToIndexMap.get(vertexTwoNeighborhood)];
	}

	/**
	 * Increases the size of the heap to support additional elements
	 */
	protected void grow()
	{
		final V[] elements = (V[]) new Object[this.array.length * 2];
		System.arraycopy(this.array, 0, elements, 0, this.array.length);
		this.array = elements;
	}

	public int indexOf(V v2)
	{
		if ( !this.elementsToIndexMap.containsKey(v2) )
			throw new IllegalArgumentException("object not contained:" + v2);
		final int index = this.elementsToIndexMap.get(v2);
		return index;
	}

	private void iniHeap(Collection<V> vertexSet)
	{
		this.heapSize = vertexSet.size();
		int t = 1;
		for (final Iterator<V> iterator = vertexSet.iterator(); iterator.hasNext();)
		{
			this.array[t] = iterator.next();
			this.elementsToIndexMap.put(this.array[t], t);
			t++;
		}
		this.ensureHeapProperty();
	}

	/**
	 * Inserts an element into queue.
	 * 
	 * @param element the element to be inserted
	 */
	public void insert(V element)
	{
		// is full?
		if ( this.array.length == this.heapSize + 1 )
		{
			this.grow();
		}

		if ( this.elementsToIndexMap.containsKey(element) )
			throw new IllegalArgumentException("element is already contained in the queue");

		this.heapSize++;
		this.array[this.heapSize] = element;
		this.elementsToIndexMap.put(element, this.heapSize);

		this.percolateUpMinHeap(this.heapSize);
	}

	/**
	 * Tests if queue is empty.
	 * 
	 * @return <code>true</code> if queue is empty; <code>false</code> otherwise.
	 */
	@Override
	public boolean isEmpty()
	{
		return this.heapSize == 0;
	}

	/**
	 * Returns an iterator over this heap's elements.
	 * 
	 * @return an iterator over this heap's elements
	 */
	@Override
	public Iterator<V> iterator()
	{
		return new Iterator<V>() {

			private int index = 1;

			private int lastReturnedIndex = -1;

			@Override
			public boolean hasNext()
			{
				return this.index <= BinaryHeap.this.heapSize;
			}

			@Override
			public V next()
			{
				if ( !this.hasNext() )
					throw new NoSuchElementException();
				this.lastReturnedIndex = this.index;
				this.index++;
				return BinaryHeap.this.array[this.lastReturnedIndex];
			}

			@Override
			public void remove()
			{
				if ( this.lastReturnedIndex == -1 )
					throw new IllegalStateException();
				BinaryHeap.this.remove(this.lastReturnedIndex);
				this.index--;
				this.lastReturnedIndex = -1;
			}

		};
	}

	@Override
	public boolean offer(V arg0)
	{
		this.insert(arg0);
		return true;
	}

	/**
	 * Returns the element on top of heap but don't remove it.
	 * 
	 * @return the element at top of heap
	 * @throws NoSuchElementException if <code>isEmpty() == true</code>
	 */
	@Override
	public V peek() throws NoSuchElementException
	{
		if ( this.isEmpty() )
			throw new NoSuchElementException();
		else
			return this.array[1];
	}

	/**
	 * Percolates element down heap from the position given by the index.
	 * <p>
	 * Assumes it is a minimum heap.
	 * 
	 * @param index the index for the element
	 */
	protected void percolateDownMinHeap(final int index)
	{
		final V element = this.array[index];
		int hole = index;

		while ((hole * 2) <= this.heapSize)
		{
			int child = hole * 2;

			// if we have a right child and that child can not be percolated
			// up then move onto other child
			if ( child != this.heapSize && this.compare(this.array[child + 1], this.array[child]) < 0 )
			{
				child++;
			}

			// if we found resting place of bubble then terminate search
			if ( this.compare(this.array[child], element) >= 0 )
			{
				break;
			}

			this.array[hole] = this.array[child];
			this.elementsToIndexMap.put(this.array[hole], hole);
			hole = child;
		}

		this.array[hole] = element;
		this.elementsToIndexMap.put(this.array[hole], hole);
	}

	/**
	 * Percolates element up heap from the position given by the index.
	 * <p>
	 * Assumes it is a minimum heap.
	 * 
	 * @param index the index of the element to be percolated up
	 */
	protected void percolateUpMinHeap(final int index)
	{
		int hole = index;
		final V element = this.array[hole];
		while (hole > 1 && this.compare(element, this.array[hole / 2]) < 0)
		{
			// save element that is being pushed down
			// as the element "bubble" is percolated up
			final int next = hole / 2;
			this.array[hole] = this.array[next];
			this.elementsToIndexMap.put(this.array[hole], hole);

			hole = next;
		}
		this.array[hole] = element;
		this.elementsToIndexMap.put(this.array[hole], hole);
	}

	/**
	 * Returns the element on top of heap and removes it.
	 * 
	 * @return the element at top of heap
	 * @throws NoSuchElementException if <code>isEmpty() == true</code>
	 */

	@Override
	public V poll()
	{
		final V result = this.peek();
		this.elementsToIndexMap.remove(result);
		this.array[1] = this.array[this.heapSize];
		this.elementsToIndexMap.put(this.array[1], 1);
		this.array[this.heapSize] = null; // for gc
		this.heapSize--;

		if ( this.heapSize != 0 )
		{// percolate top element to it's place in tree
			this.percolateDownMinHeap(1);
		}

		return result;
	}

	public void remove(int index)
	{
		this.elementsToIndexMap.remove(this.array[index]);
		this.array[index] = this.array[this.heapSize];
		this.array[this.heapSize] = null;// for gc
		this.heapSize--;

		if ( this.heapSize != 0 && index <= this.heapSize )
		{
			this.elementsToIndexMap.put(this.array[index], index);
			if ( index > 1 && this.compare(this.array[index], this.array[index / 2]) < 0 )
			{
				this.percolateUpMinHeap(index);
			}
			else
			{
				this.percolateDownMinHeap(index);
			}
		}
	}

	@Override
	public boolean remove(Object v)
	{
		final V v2 = (V) v;
		final int index = this.indexOf(v2);
		this.remove(index);
		return true;
	}

	@Override
	public int size()
	{
		return this.heapSize;
	}

	/**
	 * Returns a string representation of this heap. The returned string is similar to those
	 * produced by standard JDK collections.
	 * 
	 * @return a string representation of this heap
	 */
	@Override
	public String toString()
	{
		final StringBuffer sb = new StringBuffer();

		sb.append("[ ");

		for (int i = 1; i < this.heapSize + 1; i++)
		{
			if ( i != 1 )
			{
				sb.append(", ");
			}
			sb.append(this.array[i]);
		}

		sb.append(" ]");

		return sb.toString();
	}

	public void valueChanged(V v)
	{
		final Integer index = this.elementsToIndexMap.get(v);
		if ( index == null )
			throw new IllegalArgumentException("element not contained in the queue");

		this.percolateDownMinHeap(index);
		this.percolateUpMinHeap(index);
	}
}