package de.vibee.kanon.data;

import gnu.trove.iterator.TIntIntIterator;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

public class AdjacentGraph implements Serializable

{

	/**
	 * 
	 */
	private static final long serialVersionUID = -8468640473837439186L;

	private final List<TIntSet> graphSet;

	private final int noOfVerticesSize;

	private boolean modified;

	private int amountOfEdges = 0;

	private TIntSet[] degreeSet;

	private Map<Integer, String> indexToName;
	
	public AdjacentGraph(final List<TIntSet> graphSet, Map<Integer, String> indexToName)
	{
		this.noOfVerticesSize = graphSet.size();
		this.graphSet = graphSet;
		this.modified = false;
		this.calculateAmountOfEdges();
		this.fillDegreeSet();
		this.indexToName = indexToName;
	}
	public AdjacentGraph(final List<TIntSet> graphSet)
	{
		this(graphSet,null);
	}
	
	public String getName(int index){
		if (indexToName == null){
			return String.valueOf(index);
		}
		return indexToName.get(new Integer(index));
	}

	public boolean addEdge(final int node1, final int node2)
	{
		boolean successful = true;
		final int oldDeg1 = this.getDegree(node1);
		final int oldDeg2 = this.getDegree(node2);
		successful &= this.graphSet.get(node1).add(node2);
		successful &= this.graphSet.get(node2).add(node1);
		final int newDeg1 = this.getDegree(node1);
		final int newDeg2 = this.getDegree(node2);
		if ( successful )
		{
			this.amountOfEdges++;
			this.degreeSet[oldDeg1].remove(node1);
			this.degreeSet[oldDeg2].remove(node2);
			if ( this.degreeSet.length <= newDeg1 || this.degreeSet.length <= newDeg2 )
			{
				final TIntSet newDegreeset[] = Arrays.copyOf(this.degreeSet, this.degreeSet.length + 1);
				this.degreeSet = newDegreeset;
				this.degreeSet[this.degreeSet.length - 1] = new TIntHashSet();
			}
			this.degreeSet[newDeg1].add(node1);
			this.degreeSet[newDeg2].add(node2);
		}
		return successful;
	}

	// public boolean addEdge(int node1, int node2)
	// {
	// boolean successful = true;
	// successful &= this.graphSet.get(node1).add(node2);
	// successful &= this.graphSet.get(node2).add(node1);
	// if ( successful )
	// amountOfEdges++;
	// return successful;
	// }

	private void calculateAmountOfEdges()
	{
		this.amountOfEdges = 0;
		if ( this.graphSet != null )
		{
			for (final TIntSet adjacentNodes : this.graphSet)
			{
				if ( adjacentNodes != null )
				{
					this.amountOfEdges += adjacentNodes.size();
				}
			}
			this.amountOfEdges /= 2;
		}
	}

	private void fillDegreeSet()
	{
		this.degreeSet = new TIntSet[this.getDelta() + 1];

		for (int i = 0; i < this.degreeSet.length; i++)
		{
			this.degreeSet[i] = new TIntHashSet();
		}

		for (int i = 1; i < this.graphSet.size(); i++)
		{
			final TIntSet node = this.graphSet.get(i);
			this.degreeSet[node.size()].add(i);
		}
	}

	public TIntSet getAdjacentNodes(final int node)
	{
		return this.graphSet.get(node);
	}

	/**
	 * Return a list with all node indices that have degree n
	 * 
	 * @param n vertex degree
	 * @return Node index list
	 */
	public TIntSet getAllNodesWithDegree(final int n)
	{
		if ( n >= 0 && n < this.degreeSet.length )
			return this.degreeSet[n];
		else
			return null;
	}

	/**
	 * Return the total sum of edges in this graph
	 * 
	 * @return Amount of edges in this graph
	 */
	public int getAmountOfEdges()
	{
		return this.amountOfEdges;
	}

	public int getAmountOfNodesWithDegree(final int n)
	{
		if ( n >= 0 && n < this.degreeSet.length )
			return this.degreeSet[n].size();
		else
			return -1;
	}

	public int getAmountOfVertices()
	{
		return this.noOfVerticesSize - 1;
	}

	public int getAnonymity()
	{
		final TIntIntMap anon = new TIntIntHashMap();
		for (final TIntSet adjacentNodes : this.graphSet)
		{
			if ( adjacentNodes != null )
			{
				if ( anon.containsKey(adjacentNodes.size()) )
				{
					anon.put(adjacentNodes.size(), anon.get(adjacentNodes.size()) + 1);
				}
				else
				{
					anon.put(adjacentNodes.size(), 1);
				}
			}

		}
		int min = Integer.MAX_VALUE;
		final TIntIntIterator it = anon.iterator();
		while (it.hasNext())
		{
			it.advance();
			if ( min > it.value() )
			{
				min = it.value();
			}
		}
		return min;
	}

	/**
	 * 
	 * @param node Node index
	 * @return Degree of this node
	 */
	public int getDegree(final int node)
	{
		return this.graphSet.get(node).size();
	}

	/**
	 * Calculates the maximum vertex degree of this graph
	 * 
	 * @return Maximum vertex degree
	 */
	public int getDelta()
	{
		int delta = 0;
		for (final TIntSet node : this.graphSet)
		{
			if ( node != null )
			{
				delta = Math.max(delta, node.size());
			}
		}
		return delta;
	}

	public boolean isAdjacent(final int node1, final int node2)
	{
		return this.graphSet.get(node1).contains(node2);
	}

	public boolean isModified()
	{
		return this.modified;
	}

	public boolean removeEdge(final int node1, final int node2)
	{
		boolean successful = true;
		final int oldDeg1 = this.getDegree(node1);
		final int oldDeg2 = this.getDegree(node2);
		successful &= this.graphSet.get(node1).remove(node2);
		successful &= this.graphSet.get(node2).remove(node1);
		final int newDeg1 = this.getDegree(node1);
		final int newDeg2 = this.getDegree(node2);
		if ( successful )
		{
			this.amountOfEdges--;
			this.degreeSet[oldDeg1].remove(node1);
			this.degreeSet[oldDeg2].remove(node2);
			this.degreeSet[newDeg1].add(node1);
			this.degreeSet[newDeg2].add(node2);
		}
		return successful;
	}

	public void setModified(final boolean modified)
	{
		this.modified = modified;
	}

	// public boolean removeEdge(int node1, int node2)
	// {
	// boolean successful = true;
	// successful &= this.graphSet.get(node1).remove(node2);
	// successful &= this.graphSet.get(node2).remove(node1);
	// if ( successful )
	// amountOfEdges--;
	// return successful;
	// }

	@Override
	public String toString()
	{
		return "AdjacentGraph [size=" + this.noOfVerticesSize + ",amountOfEdges=" + this.getAmountOfEdges() + ", degreeSet="
				+ Arrays.toString(this.degreeSet) + "]";
	}

}
